"""WindowTSCoprocessor tests"""

import pandas as pd

from pylotable.win_coprocessor import WindowPandasDfGroupCoprocessor, WindowPandasDfMergeCoprocessor


def test_reference_fixture(ref_pd_df):
    """Tests the reference fixture"""

    exp_ref = pd.DataFrame()
    exp_ref['tsid'] = pd.Series(data=[0, 2, 1, 1, 1, 2])
    exp_ref['date'] = pd.Series(data=[pd.to_datetime(e) for e in [
        '2015-01-03 16:15:38',
        '2015-01-24 05:14:54',
        '2015-01-20 15:17:29',
        '2015-01-14 03:59:32',
        '2015-01-13 23:46:15',
        '2015-01-26 18:11:25']])
    pd.testing.assert_frame_equal(ref_pd_df, exp_ref)


def test_modelisation_fixture(model_pd_df):
    """Tests the modelisation fixture"""

    exp_model = pd.DataFrame()
    exp_model['tsid'] = pd.Series(data=[1, 1, 2, 2, 0, 0])
    exp_model['date'] = pd.Series(data=[pd.to_datetime(e) for e in [
        '2015-01-15 04:41:45',
        '2015-01-16 08:30:41',
        '2015-01-23 15:43:14',
        '2015-01-29 12:20:01',
        '2015-01-02 01:05:37',
        '2015-01-05 07:47:41']])
    pd.testing.assert_frame_equal(model_pd_df, exp_model)


def test_win_group_coprocessor(ref_pd_df: pd.DataFrame, model_pd_df: pd.DataFrame):
    """Tests the default window counts are correctly computed."""

    coprocessor = WindowPandasDfGroupCoprocessor.from_day_window(left_labels=('tsid', 'date'),
                                                                 right_labels=('tsid', 'date'),
                                                                 windows={
                                                                     'w_10_10': (10, 10),
                                                                     'w_1_5': (1, 5),
                                                                 })

    evaluation = (coprocessor.compute(left=ref_pd_df, right=model_pd_df)
                  .sort_values(by=[coprocessor.left_sid_label(), coprocessor.left_data_label()],
                               axis=0,
                               ascending=True))

    # sort date
    exp_eval = pd.DataFrame(index=[0, 4, 3, 2, 1, 5], data={
        'tsid': [0, 1, 1, 1, 2, 2],
        'date': [pd.to_datetime(e) for e in [
            '2015-01-03 16:15:38',
            '2015-01-13 23:46:15',
            '2015-01-14 03:59:32',
            '2015-01-20 15:17:29',
            '2015-01-24 05:14:54',
            '2015-01-26 18:11:25']],
        'w_10_10': [2, 2, 2, 2, 2, 2],
        'w_1_5': [1, 2, 2, 0, 1, 1]
    })

    pd.testing.assert_frame_equal(evaluation, exp_eval)


def test_win_merge_coprocessor(ref_pd_df: pd.DataFrame, model_pd_df: pd.DataFrame):
    """Tests the default window counts are correctly computed."""

    ref_pd_df = ref_pd_df.rename(columns={'date': 'date_ref'}, inplace=False)
    model_pd_df = model_pd_df.rename(columns={'date': 'date_mod'}, inplace=False)

    coprocessor = WindowPandasDfMergeCoprocessor.from_day_window(left_labels=('tsid', 'date_ref'),
                                                                 right_labels=('tsid', 'date_mod'),
                                                                 windows={
                                                                     'w_10_10': (10, 10),
                                                                     'w_1_5': (1, 5),
                                                                 })

    evaluation = coprocessor.compute(left=ref_pd_df, right=model_pd_df)

    # index is
    exp_eval = pd.DataFrame(index=[0, 1, 2, 3, 4, 5], data={
        'tsid': [0, 1, 1, 1, 2, 2],
        'date_ref': [pd.to_datetime(e) for e in [
            '2015-01-03 16:15:38',
            '2015-01-13 23:46:15',
            '2015-01-14 03:59:32',
            '2015-01-20 15:17:29',
            '2015-01-24 05:14:54',
            '2015-01-26 18:11:25']],
        'w_10_10': [2, 2, 2, 2, 2, 2],
        'w_1_5': [1, 2, 2, 0, 1, 1]
    })

    pd.testing.assert_frame_equal(evaluation, exp_eval)



"""A time series comparison/evaluation processing."""

from logging import getLogger

import pandas as pd

_LOG = getLogger(__name__)



class PandasDfGroupCoprocessor:
    """Co-processes two series collections here called "right" and "left"."""

    def left_sid_label(self) -> str:
        """The series id label of the left collection."""

    def left_data_label(self) -> str:
        """The data column label of the left collection."""

    def right_sid_label(self) -> str:
        """The series id label of the right collection."""

    def right_data_label(self) -> str:
        """The data column label of the right collection."""

    def preprocess_left(self, data: pd.DataFrame) -> pd.DataFrame:
        """Preprocesses the left data collection.

        Args:
            data (pd.DataFrame): the left data.

        Returns (pd.Series): the left data.
        """
        return data

    def preprocess_right(self, data: pd.DataFrame) -> pd.DataFrame:
        """Preprocesses the right data collection.

        Args:
            data (pd.DataFrame): the right data.

        Returns (pd.Series): the right data.
        """

        return data

    def preprocess_right_series(self, data: pd.DataFrame) -> pd.DataFrame | pd.Series:
        """The core process loops over each left data and processes it to the corresponding right data.

        Before each of these processing loops, the right data relative to the series id is isolated so to
        avoid useless comparisons between unrelated left series.

        Then, this right data subset is preprocessed in order to manipulate the lightest possible data.

        This last right subset preprocessing is the purpose of the current method.

        By default, it only returns the data series of the right subset corresponding to the current series to process.

        Args:
            data (pd.DataFrame): the right data subset of the currently processed series.

        Returns (pd.DataFrame | pd.Series): the right series useful for the core processing.
        """
        return data[self.right_data_label()]


    def compute_core(self, left_row: pd.Series, right_series: pd.DataFrame | pd.Series):
        """The elementary processing of a given series. For consistency purpose, inside this method, both left and
        right data must be related to the same series id, even if this information is not always used by the processing.

        Args:
            left_row (pd.Series): a series of the left data related to a single row
            right_series (pd.DataFrame | pd.Series): a subset of the right data related to the same
            series id of the left data row

        Returns: the method result is applied to each row of the left data subset related to a given series id,
        with the same right data series given in argument. Please refer to the
        pd.DataFrame.apply() method to adjust the current method return type to custom usages. The default behavior
        returns a dict in order to produce a dataframe whose column labels are the dict keys and the column values the
        successive associated dict values. The default dict maps the series id to its label in the left data
        and the left data value to the left data label.
        """
        return {
            self.left_sid_label(): left_row[self.right_sid_label()],
            self.left_data_label(): left_row[self.left_data_label()]
        }

    def compute(self, left: pd.DataFrame, right: pd.DataFrame) -> pd.DataFrame:
        """The global core processing.

        Only override it with caution. Prefers to override each data preprocessing steps.

        Preprocesses the left and right data. Then, loops over left series and applies the elementary core process to
        each of its rows.

        Args:
            left (pd.DataFrame): the left data collection; be careful to make a defensive copy before passing it as
            an argument or when overriding the preprocessing stage if no modification is wanted on the raw dataframe
            right (pd.DataFrame): the right data collection; be careful to make a defensive copy before passing it as
            an argument or when overriding the preprocessing stage if no modification is wanted on the raw dataframe

        Returns (list[pd.DataFrame]): the data computation for the whole series collections.
        """
        _LOG.debug("preprocess left data")
        left = self.preprocess_left(data=left)

        _LOG.debug("preprocess right data")
        right = self.preprocess_right(data=right)

        _LOG.debug("process group analysis")
        l = []
        for sid, left_series in left.groupby(self.left_sid_label()):

            right_data = self.preprocess_right_series(data=right[right[self.right_sid_label()] == sid])

            l.append(left_series.apply(self.compute_core,
                                        axis=1,
                                        result_type='expand',
                                        right_series=right_data))
        _LOG.debug("end of processing")
        return pd.concat(l)


class PandasDfMergeCoprocessor:
    """Co-processes two series collections here called "right" and "left"."""

    def left_sid_label(self) -> str:
        """The series id label of the left collection."""

    def left_data_label(self) -> str:
        """The data column label of the left collection."""

    def right_sid_label(self) -> str:
        """The series id label of the right collection."""

    def right_data_label(self) -> str:
        """The data column label of the right collection."""

    def preprocess_left(self, data: pd.DataFrame) -> pd.DataFrame:
        """Preprocesses the left data collection.

        Args:
            data (pd.DataFrame): the left data.

        Returns (pd.Series): the left data.
        """
        return data

    def preprocess_right(self, data: pd.DataFrame) -> pd.DataFrame:
        """Preprocesses the right data collection.

        Args:
            data (pd.DataFrame): the right data.

        Returns (pd.Series): the right data.
        """

        return data

    def merge(self, left: pd.DataFrame, right: pd.DataFrame):
        """Merges left and right series collections."""
        return pd.merge(left=left,
                        right=right,
                        how='inner',
                        left_on=[self.left_sid_label()],
                        right_on=[self.right_sid_label()],
                        validate='many_to_many')

    def preprocess(self, merge: pd.DataFrame):
        """Preprocesses the merged data before the core process applied by series."""
        return merge

    def compute_core(self, merge_series: pd.DataFrame) -> pd.DataFrame:
        """The default core processing for an individual serries."""
        return merge_series

    def process(self, merge: pd.DataFrame) -> pd.DataFrame:
        """The default core processing series by series.

        Args:
            merge (pd.DataFrame): a series of the left data inner joined to the right data for a given series.
        """
        l = []
        for _, merge_series in merge.groupby(self.left_sid_label()):
            l.append(self.compute_core(merge_series=merge_series))

        return pd.concat(l)

    def postprocess(self, merge: pd.DataFrame):
        """Postprocesses the concatenated processed data."""
        return merge

    def compute(self, left: pd.DataFrame, right: pd.DataFrame) -> pd.DataFrame:
        """The global core processing.

        Only override it with caution. Prefers to override each data preparation and preprocessing steps.

        Prepares and preprocesses the reference and modelisation data. Then, loops over reference timeseries and applies
        the elementary core process to each of its rows.

        Args:
            left (pd.DataFrame): the reference data ; be careful to make a defensive copy before passing it as
            an argument or when overriding the preparation stage if no modification is wanted on the raw dataframe
            right (pd.DataFrame): the modelisation data ; be careful to make a defensive copy before passing
            it as an argument or when overriding the preparation stage if no modification is wanted on the raw dataframe

        Returns (list[pd.DataFrame]): a list of resulting data computations for each timeseries.
        """
        _LOG.debug("preprocess left data")
        left = self.preprocess_left(data=left)

        _LOG.debug("preprocess right data")
        right = self.preprocess_right(data=right)

        _LOG.debug("merge data")
        m = self.merge(left=left, right=right)

        _LOG.debug("preprocess data")
        m = self.preprocess(m)

        _LOG.debug("process group analysis")
        m = self.process(m)

        _LOG.debug("end of processing")
        return self.postprocess(merge=m)



"""A window processing evaluation on timeseries."""

from datetime import timedelta
from logging import getLogger, DEBUG
from typing import override
from collections.abc import Hashable

import pandas as pd

from pylotable.coprocessor import PandasDfGroupCoprocessor, PandasDfMergeCoprocessor

_LOG = getLogger(__name__)

_TRACE = DEBUG - DEBUG // 2

class WindowPandasDfGroupCoprocessor(PandasDfGroupCoprocessor):
    """A window processing on series."""

    def __init__(self,
                 left_labels: tuple[str, str],
                 right_labels: tuple[str, str],
                 windows: dict[Hashable: tuple[timedelta, timedelta]]):
        self._left_sid_label = left_labels[0]
        self._left_data_label = left_labels[1]
        self._right_sid_label = right_labels[0]
        self._right_data_label = right_labels[1]
        self._windows = windows

    @override
    def left_sid_label(self) -> str:
        return self._left_sid_label

    @override
    def left_data_label(self) -> str:
        return self._left_data_label

    @override
    def right_sid_label(self) -> str:
        return self._right_sid_label

    @override
    def right_data_label(self) -> str:
        return self._right_data_label

    def preprocess_left(self, data: pd.DataFrame) -> pd.DataFrame:
        """Computes the time windows around each reference event."""

        _LOG.log(level=_TRACE, msg='compute observation / validation windows')
        data = super().preprocess_left(data=data)
        for w in self._windows:
            data[f'{w}_inf'] = data[self.left_data_label()] - self._windows[w][0]
            data[f'{w}_sup'] = data[self.left_data_label()] + self._windows[w][1]
        return data

    @override
    def compute_core(self, left_row: pd.Series, right_series: pd.DataFrame | pd.Series):
        """Counts the modelisation data included in each time window."""

        result = super().compute_core(left_row=left_row, right_series=right_series)

        _LOG.log(level=_TRACE, msg='compute observed / validated')

        for w in self._windows:
            result[w] = right_series.between(left_row[f'{w}_inf'], left_row[f'{w}_sup']).sum()
        return result

    @classmethod
    def from_day_window(cls,
                        left_labels: tuple[str, str],
                        right_labels: tuple[str, str],
                        windows: dict[Hashable, tuple[int, int]]):
        """Get a window evaluation defined by daily margins around reference events."""

        return cls(left_labels=left_labels,
                   right_labels=right_labels,
                   windows={
                       w: tuple(timedelta(days=t) for t in windows[w]) for w in windows
                   })


class WindowPandasDfMergeCoprocessor(PandasDfMergeCoprocessor):
    """A window processing on series."""

    def __init__(self,
                 left_labels: tuple[str, str],
                 right_labels: tuple[str, str],
                 windows: dict[Hashable: tuple[timedelta, timedelta]]):
        self._left_sid_label = left_labels[0]
        self._left_data_label = left_labels[1]
        self._right_sid_label = right_labels[0]
        self._right_data_label = right_labels[1]
        self._windows = windows

    @override
    def left_sid_label(self) -> str:
        return self._left_sid_label

    @override
    def left_data_label(self) -> str:
        return self._left_data_label

    @override
    def right_sid_label(self) -> str:
        return self._right_sid_label

    @override
    def right_data_label(self) -> str:
        return self._right_data_label

    @override
    def process(self, merge: pd.DataFrame) -> pd.DataFrame:
        """Computes the time windows around each reference event."""

        _LOG.log(level=_TRACE, msg='compute o')
        for w in self._windows:
            merge[w] = ((merge[self.right_data_label()] > merge[self.left_data_label()] - self._windows[w][0])
                         & (merge[self.right_data_label()] < merge[self.left_data_label()] + self._windows[w][1]))
        merge.drop(columns=[self.right_data_label()], inplace=True)
        return merge.groupby(by=[self.left_sid_label(), self.left_data_label()], as_index=False).sum()

    @classmethod
    def from_day_window(cls,
                        left_labels: tuple[str, str],
                        right_labels: tuple[str, str],
                        windows: dict[Hashable, tuple[int, int]]):
        """Get a window evaluation defined by daily margins around reference events."""

        return cls(left_labels=left_labels,
                   right_labels=right_labels,
                   windows={
                       w: tuple(timedelta(days=t) for t in windows[w]) for w in windows
                   })
